/*---------------- START OF USER DEFINABLE SETTINGS --------------------*/
/* Below are all the definitions that you will need to modify */
/* that usually dont change. Rest of them are in the config file */

/* Pretty standard command prompt, unless you've changed it		*/
#define NS_COM_PROMPT           "d> "

/* Path to the sendmail program to send out notifications               */
#define PATHTO  "/usr/sbin/sendmail -t"

/* Who do we set the from address to in an email or page?               */
#define FROM    "nobody@yourdomain.com"

/* Who to page. Normally an email address that pages an admin via modem */
/* this goes along with P_SUBJECT which I set to blank for paging       */
/* blank to disable                                                     */
#define PAGETO  "admin-pager@yourdomain.com"

/* First person to email notification with a subject of M_SUBJECT       */
/* blank to disable                                                     */
#define MAILTO1 "admin@yourdomain.com" /* who to email1 */

/* Second person to email notification with a subject of M_SUBJECT      */
/* blank to disable                                                     */
#define MAILTO2 "admin2@yourdomain.com"

/* Subject of page, if you want one                                     */
#define P_SUBJECT       ""

/* Subject of email, if you want one                                    */
#define M_SUBJECT       "NETSERVER PORT STATUS"

/* Location of config file */
#define CONF "/etc/usrchk.cfg"

/* If you want to do some debugging             */
/* valid values are 1 or 2. 2 being more info   */
#define DEBUG 0

/*---------------- END OF USER DEFINABLE SETTINGS ---------------------*/



#define MAXSES 250

/* New defines for NETSERVER modem lines of the form:
          1         2         3         4         5         6
0123456789012345678901234567890123456789012345678901234567890123456789012345678
9
S5    gambino         205.219.3.47     Netwrk  In  ESTABLISHED    11:40      0
*/

#define LINELEN  500
#define IDLE_POS  51
#define S0         0
#define LOG_POS   38
#define USER_POS   6
#define START_POS 65

#if defined(__sun__)
 /* SunOS of some sort */
  #if defined(__svr4__)
  /* Solaris, i.e. SunOS 5.x */
   #define SOL_SYS
 #else
  /* __sun__ && !__svr4__ (SunOS 4.x presumeably) */
  #define SUN4_SYS
 #endif /* __svr4__ */
#endif /* __sun__ */
#if defined(SUN4_SYS) || defined(SOL_SYS)
#define TELNET_EXEC     "/usr/ucb/telnet"
#define TELNET_CLOSE    "Connection closed"
#define TELNET_PORT     "23"
#define TELNET_PROMPT   "telnet> "
#define HAVE_SYS_WAIT
#elif defined(DNIX) || defined(HPUX)
#define TELNET_EXEC     "/usr/bin/telnet"
#define TELNET_CLOSE    "Trying"
#define TELNET_PORT     "23"
#define TELNET_PROMPT   "telnet> "
#define HAVE_SYS_WAIT
#elif defined(__linux__)
#define TELNET_EXEC     "/usr/bin/telnet"
#define TELNET_CLOSE    "Connection closed"
#define TELNET_PORT     "23"
#define TELNET_PROMPT   "telnet> "
#define HAVE_SYS_WAIT
#else /* generic entry */
#define TELNET_EXEC     "/usr/bin/telnet"
#define TELNET_CLOSE    "Connection closed"
#define TELNET_PORT     "23"
#define TELNET_PROMPT   "telnet> "
/* #define HAVE_SYS_WAIT */
/* #define HAVE_WAIT */
#endif
/* header files */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#if defined(HPUX)
#include <unistd.h>
#define NSIG _NSIG
#endif
#include <sys/types.h>
#include <ctype.h>
#if defined(HAVE_SYS_WAIT)
#include <sys/wait.h>
#elif defined(HAVE_WAIT)
#include <wait.h>
#else
extern pid_t wait(int *status);
#endif
#include <errno.h>
#include <signal.h>
#include <time.h>

/* Program declarations */
void pmwho(void);
void sigalrm(int signo);
void sigfunc(int signo);
void err_msg(const char *format, ...);
void err_exit(const char *format, ...);
void stop_telnet(void);
void midcpy(char *strf, char *strt, int fr, int to);
void remove_first(char *inpstr);
void read_config();
int  main(int argc, char *argv[]);
int  do_read(int fd, char *text1, char *text2, int show, int timeout);
int  do_write(int fd, char *text, int timeout);
char *get_time_as_string(void);
void notify_admin(char *text);

int total_hosts=0;
int globhold=0;
int dpi=0;
int MAXCHAS=0;
char *progname;
char badline[256];
char globstatus[10];
pid_t   pid;
int     got_alarm;
int     startportchecked=0;
char    mess[500];

struct {
  char groupname[32];   /* location group this chassis is in          */
  char host[32];        /* ip                                         */
  char hostname[32];    /* ip and location designation letter, if     */
                        /* 1 chassis is used for several cities       */
                        /* the letter breaks the city into a seperate */
                        /* table and makes the identifier unique      */
  char username[32];    /* username to log in to the chassis          */
  char password[32];    /* password for said username                 */
  char name[32];        /* Table header if output is in HTML          */
  int  linestotal;      /* how many total available lines on this box */
  int  beforekill;      /* how many lines should be active before we start kill */
  char startport[15];   /* for hiperarc, the slot we're in to search */
  char lastport[15];    /* up to what port/slot do we want to check? */
                        /* for hiperarc this is the slot immediately */
                        /* after the last slot we want to check  i.e. */
                        /* to check up to including slot 15, set the */
                        /* lastport at slot:16                       */
  int  sortTime;        /* 0 - sort by port, ascending         */
                        /* 1 - sort by session time, ascending */
  int  type;            /* 0 - netserver        1 - hiperarc */
} pwtable[20];


int main(int argc, char *argv[])
{
int i=0,n;

progname = argv[0];

  /* setup signals properly */
  for(n = 1; n < NSIG; n++) signal(n, sigfunc);
#if defined(SIGCHLD)
  signal(SIGCHLD, SIG_IGN);
#endif
#if defined(SIGCLD)
  signal(SIGCLD, SIG_IGN);
#endif

read_config();

for (i=0;i<MAXCHAS-1;++i) {
dpi=globhold=i;
	strcpy(pwtable[globhold].host,pwtable[globhold].hostname);
        if (isalpha(pwtable[globhold].host[strlen(pwtable[globhold].host)-1]))
           pwtable[globhold].host[strlen(pwtable[globhold].host)-1]=0;

/* we only check netserver type chassises */
if (pwtable[globhold].type==0) {
  if (DEBUG==-1) printf("Running pmwho on %s\n",pwtable[globhold].host);
  pmwho();
  }
} /* end of for */
return 1;
} /* end of main */


/* "show all" and logging out from 'hostname'              */
void pmwho()
{
  int fdpipe[2][2], fdin, fdout, fderr, err, n;
  char pw[50];
  char username2[50];

  pw[0]=0;
  username2[0]=0;
  got_alarm = 0;

  /* we need pipes to communicate between the programs */
  if (pipe(fdpipe[0]) < 0) {
    err_msg("%s: pipe failed: errno=%d (%s)\n", progname, errno, strerror(errno));
    return;
  }
  if (pipe(fdpipe[1]) < 0) {
    err_msg("%s: pipe failed: errno=%d (%s)\n", progname, errno, strerror(errno));
    close(fdpipe[0][0]);
    close(fdpipe[0][1]);
    return;
  }

  switch(pid = fork()) {
    case 0: /* child */
      /* child:stdin */
      close(0);
      if (dup(fdpipe[0][0]) < 0) {
        err_exit("%s: dup failed: errno=%d (%s)\n", progname, errno, strerror(errno));
      }

      /* close first pipe */
      close(fdpipe[0][0]);
      close(fdpipe[0][1]);

      /* child:stdout */
      close(1);
      if (dup(fdpipe[1][1]) < 0) {
        err_exit("%s: dup failed: errno=%d (%s)\n", progname, errno, strerror(errno));
      }

      /* child:stderr */
      if ((fderr = dup(2)) < 0) {
        err_exit("%s: dup failed: errno=%d (%s)\n", progname, errno, strerror(errno));
      }

      close(2);
      if (dup(fdpipe[1][1]) < 0) {
        err = errno;
        dup(fderr);
        err_exit("%s: dup failed: errno=%d (%s)\n", progname, err, strerror(err));
      }

      /* close second pipe */
      close(fdpipe[1][0]);
      close(fdpipe[1][1]);

      /* exec TELNET application */
      execl(TELNET_EXEC, "telnet", pwtable[globhold].host, TELNET_PORT, (char*)NULL);

      /* if we're still here the TELNET_EXEC could not be exec'd */
      err = errno;
      close(2);
      dup(fderr);
      err_exit("%s: tt: execl(%s) failed to %s:%s errno=%d (%s)\n", progname, pwtable[globhold].host, TELNET_PORT, TELNET_EXEC, err, strerror(err));
      break;

    case -1: /* error */
      err_exit("%s: fork failed: errno=%d (%s)\n", progname, errno, strerror(errno));
      return;

    default: /* parent */
      /* close the childs end of the pipes */
      close(fdpipe[0][0]);
      close(fdpipe[1][1]);
      break;
  }

  /* now communicate with the 'telnet' process */
  fdin = fdpipe[1][0];
  fdout = fdpipe[0][1];

strcpy(pw,pwtable[globhold].password);
strcpy(username2,pwtable[globhold].username);
badline[0]=0;

  for(;;) {
    n = do_read(fdin, "login: ", TELNET_PROMPT, 0, 15);
    if (n != 1) {
      n = -1; /* TELNET_PROMPT == connection failed */
      break;
    } /* end of if */

	if ((n = do_write(fdout, username2, strlen(username2))) < 0) break;

	if ((n = do_write(fdout, "\n", 1)) < 0) break;

    if ((n = do_read(fdin, "Password: ", NULL, 0, 10)) < 0) break;


	if ((n = do_write(fdout, pw, strlen(pw))) < 0) break;
	if ((n = do_write(fdout, "\n", 1)) < 0) break;

    if (n < 0) break;

	if ((n = do_write(fdout, "\n", 1)) < 0) break;

    if ((n = do_read(fdin, NS_COM_PROMPT, NULL, 0, 10)) < 0) break;

    if ((n = do_write(fdout, "show all\n", 5)) < 0) break;

    for(;;) {

      n = do_read(fdin, "-- Press Return for More -- ", NS_COM_PROMPT, 1, 30);
      if ((n < 0) || (n == 2)) break;

      if ((n = do_write(fdout, "\n", 5)) < 0) break;

    } /* end of sub for */
    if (n < 0) break;

startportchecked=0;

	if ((n = do_write(fdout, "quit\n", 5)) < 0) break;
    if ((n = do_read(fdin, TELNET_CLOSE, NULL, 0, 20)) < 0) break;

    break;
  } /* end of forever for */

  close(fdin);
  close(fdout);

  if ((n < 0) && (got_alarm == 0)) {
    err_msg("%s: connection to host '%s' failed\n", progname, pwtable[globhold].host);
  }

  stop_telnet();

/* there is bad ports */
if (strlen(badline)) {
sprintf(mess,"Bad ports on host %s, stat %s, %s",
pwtable[dpi].hostname,globstatus,badline);
notify_admin(mess);
} /* end of if badline */

}

/* read information from 'telnet', stop reading upon errors, on a timeout */
/* and when the beginning of a line equals to 'text1' (and then return 1) */
/* or when the beginning of a line equals to 'text2' (and then return 2), */
/* if 'show' is non zero we display the information to the screen         */
/* (using 'stdout').                                                      */
int do_read(int fd, char *text1, char *text2, int show, int timeout)
{
  char line[LINELEN+1];
  char port[10];
  char status[10];
  int n, len1, len2, m, t;
  char str1[20];

line[0]=0;
port[0]=0;
status[0]=0;

  /* setup alarm (so we won't hang forever upon problems) */
  signal(SIGALRM, sigalrm);
  alarm(timeout);

  len1 = strlen(text1);
  len2 = (text2 != NULL) ? strlen(text2) : 0;

  /* start reading from 'telnet' */
  for(;;) {
    n = 0;
	for(t=0;t<LINELEN+1;t++)line[t]='\0';t=0;
	/* line[0]=0; */
    for(;;) {
      if (n == LINELEN) {
        alarm(0); /* disable alarm */
        stop_telnet();
        err_exit("%s: too long line!\n", progname);
      } /* end of if n */
      m = read(fd, &line[n], 1);
      if (m == -1) {
	sprintf(mess,"%s: read() Error in do_read%s: %s\n",get_time_as_string(),"",strerror(errno));
	printf("%s",mess);
	sprintf(mess,"%s: Line read: %s\n",get_time_as_string(),line);
	printf("%s",mess);

        alarm(0); /* disable alarm */
        return(-1);
      } /* end of if m */
      if ((line[n] == '\r')) continue;
      if (n >= len1-1) {
        if (strncmp(&line[n-(len1-1)], text1, len1) == 0) {
          /* we found the keyword we were searching for */
          alarm(0); /* disable alarm */
          return(1);
        }
      } /* end of if n */
      if ((text2 != NULL) && (n >= len2-1)) {
        if (strncmp(&line[n-(len2-1)], text2, strlen(line)) == 0) {
          /* we found the keyword we were searching for */
          alarm(0); /* disable alarm */
          return(2);
        }
      } /* end of if text2 */
      if (line[n] == '\n') break;
      n++;
    } /* end of sub for */

    if (show) {

      line[++n] = '\0';

if (startportchecked==2) {
	if (DEBUG>=1) printf("in do_read after last port 1!\n");
	fflush(stdout);
	return(1);
	}

/* Netserver chassis card */
if (pwtable[dpi].type==0) {
	/* First 2 lines of output */
	if (((line[0] == '-') && (line[1] == '-')) ||
            (strncmp(&line[0], "Port",4) == 0)){
            fflush(stdout);
        } /* end of if */
	else {

        for(t=0;t<sizeof(str1);t++) str1[t]='\0'; t=0;
	/* check for last port */
	if (startportchecked==1) {
	strncpy(str1, &line[S0],4);
	if (strstr(str1,pwtable[dpi].lastport)) {
		if (DEBUG>=1) printf("Hit last port %s\n",pwtable[dpi].lastport);
                for(t=0;t<sizeof(str1);t++)str1[t]='\0'; t=0;
		startportchecked=2;
		}
	}

	/* check for first port */
	if (!startportchecked) {
	strncpy(str1, &line[S0],4);
	if (!strstr(str1,pwtable[dpi].startport)) {
                for(t=0;t<sizeof(str1);t++)str1[t]='\0'; t=0;
		continue;
		}
	else startportchecked=1;
                for(t=0;t<sizeof(str1);t++)str1[t]='\0'; t=0;
	}

for(t=0;t<sizeof(port);t++) port[t]='\0'; t=0;
strncpy(port, &line[S0], 4);
if (strlen(port)<=1) continue;
for(t=strlen(port);t>0;--t) {
	if (isalnum((int)port[t])) {
	port[t+1]=0; break;
	}
}


for(t=0;t<sizeof(status);t++) status[t]='\0'; t=0;
strncpy(status, &line[5], 5);
if (!strcmp(status,"A R P")) continue;

if (!strlen(badline)) strcpy(globstatus,status);
if (!strlen(badline)) strcat(badline,port);
else { strcat(badline,","); strcat(badline,port); }

if (DEBUG==-1) printf("port \"%s\" status \"%s\"\n",port,status);

          } /* end of else not header line */
	} /* end of Netserver */

      } /* end of if show */
    } /* end of for */
}

void read_config()
{
int zz=0;
char fileline[257];
char readbuf[256];
char varbuf[256];
FILE *fp2;

if (!(fp2=fopen(CONF,"r")))
        {
         printf("Cant open config file!\n");
         exit(0);
        }
else {
        zz=0;
        while (fgets(fileline,256,fp2) != NULL) {
                fileline[strlen(fileline)-1]=0;
        if (fileline[0]=='#' || fileline[0]==' ' || !strlen(fileline)) continue;
        else {
        varbuf[0]=0;
        sscanf(fileline,"%s ",varbuf);
        remove_first(fileline);
        if (!strcmp(varbuf,"BOX:")) {
                /* hostname */
                sscanf(fileline,"%s ",pwtable[zz].hostname);
                remove_first(fileline);
                /* username */
                sscanf(fileline,"%s ",pwtable[zz].username);
                remove_first(fileline);
                /* password */
                sscanf(fileline,"%s ",pwtable[zz].password);
                remove_first(fileline);
                /* lines total */
                sscanf(fileline,"%s ",readbuf);
                pwtable[zz].linestotal=atoi(readbuf);
                remove_first(fileline);
                /* before kill */
                sscanf(fileline,"%s ",readbuf);
                pwtable[zz].beforekill=atoi(readbuf);
                remove_first(fileline);
                /* start port */
                sscanf(fileline,"%s ",pwtable[zz].startport);
                remove_first(fileline);
                /* last port */
                sscanf(fileline,"%s ",pwtable[zz].lastport);
                remove_first(fileline);
                /* sort time */
                sscanf(fileline,"%s ",readbuf);
                pwtable[zz].sortTime=atoi(readbuf);
                remove_first(fileline);
                /* chassis type */
                sscanf(fileline,"%s ",readbuf);
                pwtable[zz].type=atoi(readbuf);
                remove_first(fileline);
                /* group */
                sscanf(fileline,"%s ",pwtable[zz].groupname);
                remove_first(fileline);
                /* chassis description */
                strcpy(pwtable[zz].name,fileline);
                zz++;
        } /* end of BOX: else if */
        else continue;
      } /* end of else */
    } /* end of while */
}
fclose(fp2);
MAXCHAS=zz+1;
}

/* print error text */
void err_msg(const char *format, ...)
{
  va_list ap;
  /* show error message */
  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);
  fflush(stderr);
}

/* print error text and exit gracefully */
void err_exit(const char *format, ...)
{
  va_list ap;
  /* show error message */
  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);
  fflush(stderr);
  /* exit with error */
  exit(1);
}

/* try to terminate the telnet-child that we started */
void stop_telnet(void)
{
  pid_t p;
  if ((pid != -1) && (pid != 0)) {
    if (kill(pid, 0) >= 0) {
      kill(pid, SIGTERM);
      kill(pid, SIGKILL);
      if (kill(pid, 0) >= 0) {
        for(p = wait(0); (p != -1) && (p != pid); p = wait(0)) ; /* zombies */
      }
    }
    pid = -1;
  }
}


void midcpy(char *strf, char *strt, int fr, int to)
{
int f;
for (f=fr;f<=to;++f)
  {
   if (!strf[f])
     {
      strt[f-fr]='\0';
      return;
     }
   strt[f-fr]=strf[f];
  }
strt[f-fr]='\0';
}

/*** removes first word at front of string and moves rest down ***/
void remove_first(char *inpstr)
{
int newpos,oldpos;
newpos=0;  oldpos=0;
/* find first word */
while(inpstr[oldpos]==' ') {
        if (!inpstr[oldpos]) { inpstr[0]=0;  return; }
        oldpos++;
        }
/* find end of first word */
while(inpstr[oldpos]!=' ') {
        if (!inpstr[oldpos]) { inpstr[0]=0;  return; }
        oldpos++;
        }
/* find second word */
while(inpstr[oldpos]==' ') {
        if (!inpstr[oldpos]) { inpstr[0]=0;  return; }
        oldpos++;
        }
while(inpstr[oldpos]!=0)
        inpstr[newpos++]=inpstr[oldpos++];
inpstr[newpos]='\0';
}


/* write a command to the 'telnet' process */
int do_write(int fd, char *text, int timeout)
{
  int err, len, n;

  /* setup alarm (so we won't hang forever upon problems) */
  signal(SIGALRM, sigalrm);
  alarm(timeout);

  len = strlen(text);
  n = write(fd, text, len);
  if (n != len) {
    err = errno;
    alarm(0); /* disable alarm */
#if 0
    if (n < 0) {
      err_msg("%s: write failed: errno=%d (%s)\n", progname, err, strerror(err));
    }
#endif
    return(-1);
  }
  else {
    alarm(0); /* disable alarm */
    return(0);
  }
}

/* our timeout procedure, used to abort malfunctioning connections */
void sigalrm(int signo)
{
  got_alarm = 1;
  err_msg("%s: timeout on connection to host '%s'\n", progname, pwtable[globhold].host);
  stop_telnet();
  exit(1);
}

/* handle the reception of signals */
void sigfunc(int signo)
{
time_t tm;
#if defined(SIGPIPE)
  if (signo != SIGPIPE)
#endif
  {
        time(&tm);
    err_msg("%s: received signal #%d during connection to host %s --aborting: %s\n", progname, signo, pwtable[globhold].host,ctime(&tm));
  }
  stop_telnet();
  exit(1);
}

char *get_time_as_string(void)
{
    time_t now;
    char *timestring;
    time(&now);
    timestring = ctime(&now);
    timestring[strlen(timestring) - 1] = '\0'; /*trim newline at end*/

    return timestring;
}


void notify_admin(char *text)
{
FILE *pp;
char tem[256];

sprintf(tem,"%s",PATHTO);

if (strlen(PAGETO)) {
  if (!(pp=popen(tem,"w")))
    {
     exit(0);
    }
   fprintf(pp,"From: \"%s\"\n",FROM);
   fprintf(pp,"To: %s\n",PAGETO);
   fprintf(pp,"Subject: %s\n\n",P_SUBJECT);
   fprintf(pp,"%s",text);
   fputs("\n\n",pp);
   fputs(".\n",pp);
   pclose(pp);
}

if (strlen(MAILTO1)) {
  if (!(pp=popen(tem,"w")))
    {
     exit(0);
    }
   fprintf(pp,"From: \"%s\"\n",FROM);
   fprintf(pp,"To: %s\n",MAILTO1);
   fprintf(pp,"Subject: %s\n\n",M_SUBJECT);
   fprintf(pp,"%s",text);
   fputs("\n\n",pp);
   fputs(".\n",pp);
   pclose(pp);
}

if (strlen(MAILTO2)) {
  if (!(pp=popen(tem,"w")))
    {
     exit(0);
    }
   fprintf(pp,"From: \"%s\"\n",FROM);
   fprintf(pp,"To: %s\n",MAILTO2);
   fprintf(pp,"Subject: %s\n\n",M_SUBJECT);
   fprintf(pp,"%s",text);
   fputs("\n\n",pp);
   fputs(".\n",pp);
   pclose(pp);
}

}

